import os
from tkinter import OptionMenu
import numpy as np
import scipy.optimize
from itertools import combinations

if os.name == 'nt':
    from kullback_leibler import klucbBern, klucbGauss, klGauss
else:
    import pyximport
    # print('used')
    _ = pyximport.install()
    from kullback_leibler_cython import klucbBern, klucbGauss, klGauss


def kl_ucb(n_list, emp_avg_list, t, type='Bern'):
    kl_ucb_list = np.zeros(len(n_list))
    f_t = np.log(t) + 4 * np.log(np.log(t))
    for k in range(len(n_list)):
        if n_list[k] > 0:
            if type == 'Bern':
                kl_ucb_list[k] = klucbBern(emp_avg_list[k], f_t/n_list[k])
            else:
                kl_ucb_list[k] = klucbGauss(
                    emp_avg_list[k], f_t/n_list[k], sig2x=0.5)
        else:
            kl_ucb_list[k] = np.inf

    return kl_ucb_list


class MPMABSA:
    '''
    MAB with arm sharing model:

    Inputs:
    =======================================================
    N: the number of players
    K: the number of arms
    mu_list: a 1xK list, the reward mean of each arm;
    m_list: a 1xK list, the number of cores for each arm;
    rw_dist: reward distribution, 'normal' or 'Bernoulli'
    '''

    def __init__(self, N, K, mu_list, m_list, rw_dist='Bernoulli', alpha=1, beta=0):
        # number of players
        self.N = N
        # number of arms
        self.K = K
        # reward mean
        assert(len(mu_list) == self.K)
        self.mu_list = mu_list
        # each arm cores number
        assert(len(m_list) == self.K)
        self.m_list = m_list
        # reward distribution
        self.rw_dist = rw_dist

        # beta and alpha function
        self.alpha = alpha
        self.beta = beta

        # oracle reward and player allocation
        self.oc_alc, self.oc_rw = oracle(
            self.K, self.N, self.mu_list, self.m_list, self.alpha, self.beta)


    def general_scaling(self, min_a_m):
        if min_a_m == 0:
            return 0
        else:
            return min_a_m**self.alpha + self.beta

    def single_arm_pull(self, i, j):
        '''
        j players share arm i:

        Inputs:
        ==========================================================
        i\in [K], the arm's index
        j\in [N], the number of players select arm i at the same time

        Notice: only one effective sample, under the arm level feedback.
        Outputs:
        ==========================================================
        a 1xj list, its kth entry is the reward for the kth player
        '''
        if j == 0:
            return 0
        else:
            rw_sample = 0
            np.random.seed(int.from_bytes(os.urandom(4), byteorder='little'))
            if self.rw_dist == 'normal':
                rw_sample = np.random.normal(
                    loc=self.mu_list[i], scale=1/np.sqrt(2), size=1)
            elif self.rw_dist == 'Bernoulli':
                rw_sample = np.random.binomial(n=1, p=self.mu_list[i], size=1)
            else:
                rw_sample = self.mu_list[i]

            return rw_sample * self.general_scaling(min(j, self.m_list[i]))

    def combinatorial_arm_pull(self, player_list):
        '''
        player distribution to K arms

        Input:
        ==============================================================
        player_list: a 1xK list, its kth entry represent that the number of
        player (in the entry) is playing the kth arm.

        Output:
        ==============================================================
        a 1xK list for each arm's total_reward of the combintorial arm
        '''
        # print(player_list)
        # assert(sum(player_list) == self.N)
        rw_list = np.zeros(self.K)

        for i, j in zip(range(self.K), player_list):
            rw_list[i] = self.single_arm_pull(i, j)

        return rw_list


def oracle(K, N, mu_list, m_list, alpha=1, beta=0):
    '''
    Oracle Identification function

    Input:
    ================================================================
    The class itself

    Assumption:
    ================================================================
    K > N and m_i > 1 for all i\in [K]
    which implies that in the oracle distribution, no arm is busy. For
    each arm, the number of players is less than or each to its m_i.

    Output:
    ================================================================
    A 1xK list show the player distribution for orcale
    A expected total reward for the oracle (can be used for calculating
    regret).
    '''
    ranking = np.argsort(mu_list)
    # print('ranking: ', ranking)
    oc_total_rw = 0
    budget = N
    oc_alc = np.zeros(K, dtype=int)

    for i in range(K-1, -1, -1):

        if budget >= m_list[ranking[i]]:
            budget -= m_list[ranking[i]]
            oc_alc[ranking[i]] = m_list[ranking[i]]
            # print(self.mu_list)
            oc_total_rw += mu_list[ranking[i]] * \
                (m_list[ranking[i]]**alpha + beta)
        elif budget < m_list[ranking[i]] and budget > 0:
            oc_total_rw += mu_list[ranking[i]] * (budget**alpha + beta)
            oc_alc[ranking[i]] = budget
            # budget = 0
            break

    return oc_alc, oc_total_rw




def U(tau, T):
    # confidence interval for elimination
    return 2 * np.sqrt(2 * max(1, np.log(T / tau)) / tau)


def masked_del(S, a):
    # input condidate arm set S, and the arm a to eliminate
    mask = np.ones(len(S), dtype=bool)
    mask[S == a] = False
    return S[mask]


def partition(K, N):
    # return a generator for action space, size (N+K-1, K-1)
    # divide N into K
    n = K + N - 1
    for splits in combinations(range(n), K - 1):
        yield [s1 - s0 - 1 for s0, s1 in zip((-1,) + splits, splits + (n,))]


def uci(tau, iota, dlt, hat_mu, hat_nu, pre_lcb, pre_ucb):
    # input: IE sample size, UE sample size, m_ucb upper bound of secondary users, hat_mu, hat_nu = \mu*m
    # output: the lower confidence bound (ceil) and upper confidence bound (ucb)
    #                if not enough sample, lcb=1, ucb=N

    def phi(x, dlt):
        return np.sqrt((1+1/x) * np.log(2 * np.sqrt(x+1) / dlt) / (2*x))
        # return np.sqrt(np.log(2 / dlt) / (2*x))

    phi_iota = phi(iota, dlt)
    phi_tau = phi(tau, dlt)

    if hat_mu - phi_iota - phi_tau <= 0:
        return pre_lcb, pre_ucb

    return max(pre_lcb, np.ceil(hat_nu / (hat_mu + phi_tau + phi_iota))), min(pre_ucb, np.floor(hat_nu / (hat_mu - phi_tau - phi_iota)))


def uci_easy(tau, iota, dlt, hat_mu, hat_nu, N):
    # input: IE sample size, UE sample size, m_ucb upper bound of secondary users, hat_mu, hat_nu = \mu*m
    # output: the lower confidence bound (ceil) and upper confidence bound (ucb)
    #                if not enough sample, lcb=1, ucb=N

    def phi(x, dlt):
        return np.sqrt((1+1/x) * np.log(2 * np.sqrt(x+1) / dlt) / (2*x))
        # return np.sqrt(np.log(2 / dlt) / (2*x))

    phi_iota = phi(iota, dlt)
    phi_tau = phi(tau, dlt)

    if hat_mu - phi_iota - phi_tau <= 0:
        return 1, N

    return max(1, np.ceil(hat_nu / (hat_mu + phi_tau + phi_iota))), min(N, np.floor(hat_nu / (hat_mu - phi_tau - phi_iota)))


def hfd_uci(tau, iota, dlt, hat_mu, hat_nu, pre_lcb, pre_ucb):
    # input: IE sample size, UE sample size, m_ucb upper bound of secondary users, hat_mu, hat_nu = \mu*m
    # output: the lower confidence bound (ceil) and upper confidence bound (ucb)
    #                if not enough sample, lcb=1, ucb=N

    def rho(x, dlt):
        # return np.sqrt((1+1/x) * np.log(2 * np.sqrt(x+1) / dlt) / (2*x))
        return np.sqrt(np.log(2 / dlt) / (2*x))

    rho_iota = rho(iota, dlt)
    rho_tau = rho(tau, dlt)

    if hat_mu - rho_iota - rho_tau <= 0:
        return pre_lcb, pre_ucb

    return max(pre_lcb, np.ceil(hat_nu / (hat_mu + rho_tau + rho_iota))), min(pre_ucb, np.floor(hat_nu / (hat_mu - rho_tau - rho_iota)))


def MPSESA(MAB, gamma=1, xi=1, T=5000, type='uci'):

    # Initialization
    rw_T_list = np.zeros(T)

    # estimations

    hat_mu_list = np.zeros(MAB.K)
    hat_nu_list = np.zeros(MAB.K)
    m_flag = np.zeros(MAB.K)

    hat_m_list = np.ones(MAB.K, dtype=int)
    hat_m_upper = np.ones(MAB.K, dtype=int) * MAB.N

    # candidate arm set
    S = np.arange(MAB.K, dtype=int)

    # time index and sample size index
    tau = 1
    iota_list = np.ones(MAB.K, dtype=int)
    t = 1

    while t <= T:
        order = np.argsort([hat_mu_list[i] for i in S])
        # estimate L
        threshold_L = 1
        # print(hat_m_list)
        s = len(S)
        while threshold_L < s and sum([hat_m_list[S[order[-i-1]]] for i in range(threshold_L)]) < MAB.N:
            threshold_L = threshold_L + 1

        if threshold_L < len(S):
            S_prime = S
            gap = gamma * U(tau, T)

            for i in range(s-1, 0, -1):
                if hat_mu_list[S_prime[order[-threshold_L]]] - gap > hat_mu_list[S_prime[order[-i-1]]]:
                    S = masked_del(S, S_prime[order[-i-1]])
                else:
                    break

            # individually explore S and update hat_mu_list
            s = len(S)
            s_round = s // MAB.N
            s_remind = s % MAB.N

            for i in range(s_round):
                action = np.zeros(MAB.K)
                for j in range(MAB.N):
                    action[S[j + i*MAB.N]] = 1

                rw_list = MAB.combinatorial_arm_pull(action)
                rw_T_list[t-1] = sum(rw_list)

                hat_mu_list = [((tau-1) * hat_mu_list[i] + (rw_list[i]/action[i])) /
                               tau if action[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]
                t = t + 1
                if t > T:
                    return rw_T_list

            if s_remind != 0:
                action = np.zeros(MAB.K)
                for i in range(s_remind):
                    action[S[-i-1]] = 1

                add_action, _ = oracle(
                    MAB.K, MAB.N - s_remind, hat_mu_list, hat_m_list - action)

                action = action + add_action

                rw_list = MAB.combinatorial_arm_pull(action)
                rw_T_list[t-1] = sum(rw_list)

                hat_mu_list = [((tau-1) * hat_mu_list[i] + (rw_list[i]/action[i])) /
                               tau if action[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]
                t = t + 1
                if t > T:
                    return rw_T_list

            tau = tau + 1

        elif threshold_L == len(S):
            action, _ = oracle(MAB.K, MAB.N, hat_mu_list, hat_m_list)
            rw_list = MAB.combinatorial_arm_pull(action)
            rw_T_list[t-1] = sum(rw_list)

            hat_mu_list = [((tau-1) * hat_mu_list[i] + (rw_list[i]/action[i])) /
                           tau if action[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]

            t = t + 1
            if t > T:
                return rw_T_list
            tau = tau + 1

            for arm in S:
                if m_flag[arm] == 0:

                    action = np.zeros(MAB.K)
                    action[arm] = hat_m_upper[arm]

                    add_action, _ = oracle(MAB.K, MAB.N - action[arm], hat_mu_list, [
                        0 if i == arm else hat_m_list[i] for i in range(MAB.K)])
                    action = action + add_action

                    rw_list = MAB.combinatorial_arm_pull(action)
                    rw_T_list[t-1] = sum(rw_list)

                    hat_nu_list[arm] = (
                        (iota_list[arm]-1) * hat_nu_list[arm] + rw_list[arm])/iota_list[arm]

                    t = t+1
                    if t > T:
                        return rw_T_list

                    iota_list[arm] = iota_list[arm] + 1

                    if type == 'uci':
                        hat_m_list[arm], hat_m_upper[arm] = uci(
                            tau, iota_list[arm], xi/T, hat_mu_list[arm], hat_nu_list[arm], hat_m_list[arm], hat_m_upper[arm])
                    if type == 'uci_easy':
                        hat_m_list[arm], hat_m_upper[arm] = uci_easy(
                            tau, iota_list[arm], xi/T, hat_mu_list[arm], hat_nu_list[arm], MAB.N)
                    elif type == 'hfd':
                        hat_m_list[arm], hat_m_upper[arm] = hfd_uci(
                            tau, iota_list[arm], xi/(T**2), hat_mu_list[arm], hat_nu_list[arm], hat_m_list[arm], hat_m_upper[arm])

                    if hat_m_list[arm] == hat_m_upper[arm]:
                        m_flag[arm] = 1

        # if len(S) <= MAB.N:

        # print(
        #     f"candidate arm set ={S}, L={threshold_L}, time={t, tau, iota_list}, gap={gap}, hat_m_list={hat_m_list}, hat_m_upper={hat_m_upper}")

    return rw_T_list



def MP_SE(MAB, gamma=1, T=5000):
    ################################################################
    # Input: Assume Known m
    #       T: the time horizon
    # Output:
    #       a 1 x T reward list;
    ################################################################

    # Initialization
    rw_T_list = np.zeros(T)
    hat_mu_list = np.zeros(MAB.K)

    tau = 1
    S = np.arange(MAB.K, dtype=int)
    t = 1
    gma = gamma

    while t <= T:
        order = np.argsort([hat_mu_list[i] for i in S])
        s = len(S)
        S_prime = S
        current_cap = sum([MAB.m_list[i] for i in S])
        if current_cap >= MAB.N and current_cap - MAB.m_list[S_prime[order[-s]]] < MAB.N:
            action, _ = oracle(MAB.K, MAB.N, [
                hat_mu_list[i] if i in S else 0 for i in range(MAB.K)], MAB.m_list)
            rw_list = MAB.combinatorial_arm_pull(action)
            rw_T_list[t-1] = sum(rw_list)

            t = t + 1
            if t > T:
                return rw_T_list

            hat_mu_list = [((tau-1) * hat_mu_list[i] + rw_list[i]/action[i]) /
                           tau if action[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]

        else:
            gap = gma * U(tau, T)

            threshold_L = 1
            while threshold_L < s and sum([MAB.m_list[S[order[-i-1]]] for i in range(threshold_L)]) < MAB.N:
                threshold_L = threshold_L + 1

            for i in range(s-1, 0, -1):
                if hat_mu_list[S_prime[order[-threshold_L]]] - hat_mu_list[S_prime[order[-i-1]]] > gap:
                    S = masked_del(S, S_prime[order[-i-1]])
                else:
                    break

            # renew s after elimination
            s = len(S)
            s_round = s // MAB.N
            s_remind = s % MAB.N

            for i in range(s_round):
                action = np.zeros(MAB.K)
                for j in range(MAB.N):
                    action[S[j + i*MAB.N]] = 1

                rw_list = MAB.combinatorial_arm_pull(action)
                rw_T_list[t-1] = sum(rw_list)
                t = t + 1
                hat_mu_list = [((tau-1) * hat_mu_list[i] + rw_list[i]/action[i]) /
                               tau if action[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]

            if s_remind != 0:
                action = np.zeros(MAB.K)
                for i in range(s_remind):
                    action[S[-i-1]] = 1

                add_action, _ = oracle(
                    MAB.K, MAB.N - s_remind, hat_mu_list, MAB.m_list - action)

                action = action + add_action

                rw_list = MAB.combinatorial_arm_pull(action)
                rw_T_list[t-1] = sum(rw_list)
                t = t + 1
                if t > T:
                    return rw_T_list

                hat_mu_list = [((tau-1) * hat_mu_list[i] + rw_list[i]/action[i]) /
                               tau if action[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]

        tau = tau + 1

    return rw_T_list


def CUCB(MAB, gamma=1, T=5000, type='ucb'):
    ################################################################
    # Input:
    #       t0: 1-1 player - arm exploration
    #       t1: N-1 player - arm exploitation
    #       T: the time horizon
    # Output:
    #       a 1 x T reward list;
    ################################################################

    # Initialization
    rw_T_list = np.zeros(T)
    hat_mu_list = np.zeros(MAB.K)

    t = 1

    counter_ind = np.ones(MAB.K, dtype=int)

    # initial round
    s_round = MAB.K // MAB.N
    s_remind = MAB.K % MAB.N

    for i in range(s_round):
        action = np.zeros(MAB.K)
        for j in range(MAB.N):
            action[j + i*MAB.N] = 1

        rw_list = MAB.combinatorial_arm_pull(action)
        rw_T_list[t-1] = sum(rw_list)

        hat_mu_list = [((counter_ind[i]-1) * hat_mu_list[i] + rw_list[i]/action[i]) /
                       counter_ind[i] if action[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]
        counter_ind = [counter_ind[i] + 1 if action[i] >
                       0 else counter_ind[i] for i in range(MAB.K)]
        t = t + 1
        if t > T:
            return rw_T_list

    if s_remind != 0:
        action = np.zeros(MAB.K)
        for i in range(s_remind):
            action[-i-1] = 1

        add_action, _ = oracle(MAB.K, MAB.N - s_remind,
                               hat_mu_list, MAB.m_list - action)

        action = action + add_action

        rw_list = MAB.combinatorial_arm_pull(action)
        rw_T_list[t-1] = sum(rw_list)

        hat_mu_list = [((counter_ind[i]-1) * hat_mu_list[i] + rw_list[i]/action[i]) /
                       counter_ind[i] if action[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]
        counter_ind = [counter_ind[i] + 1 if action[i] >
                       0 else counter_ind[i] for i in range(MAB.K)]

        t = t + 1
        if t > T:
            return rw_T_list

    while t <= T:

        # UCB Index:
        if type == 'ucb':
            ucb_mu_list = hat_mu_list + gamma * \
                np.sqrt(1.5 * np.log(t) / counter_ind)
        elif type == 'klucb':
            ucb_mu_list = kl_ucb(counter_ind, hat_mu_list, t)

        # decision
        action, _ = oracle(MAB.K, MAB.N, ucb_mu_list, MAB.m_list)

        # play
        rw_list = MAB.combinatorial_arm_pull(action)
        rw_T_list[t-1] = np.sum(rw_list)

        # record and update
        hat_mu_list = [((counter_ind[i]-1) * hat_mu_list[i] + rw_list[i]/min(MAB.m_list[i], action[i])
                        )/counter_ind[i] if action[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]
        counter_ind = [counter_ind[i] + 1 if action[i] >
                       0 else counter_ind[i] for i in range(MAB.K)]

        t = t + 1

    return rw_T_list


def CUCB1(MAB, gamma=1, T=5000, type='ucb', isawsexplore=True):
    ################################################################
    # Input:
    #       t0: 1-1 player - arm exploration
    #       t1: N-1 player - arm exploitation
    #       T: the time horizon
    # Output:
    #       a 1 x T reward list;
    ################################################################

    # Initialization
    rw_T_list = np.zeros(T)
    hat_mu_list = np.zeros(MAB.K)

    t = 1

    counter_ind = np.ones(MAB.K, dtype=int)

    # initial round
    s_round = MAB.K // MAB.N
    s_remind = MAB.K % MAB.N

    for i in range(s_round):
        action = np.zeros(MAB.K)
        for j in range(MAB.N):
            action[j + i*MAB.N] = 1

        rw_list = MAB.combinatorial_arm_pull(action)
        rw_T_list[t-1] = sum(rw_list)

        hat_mu_list = [((counter_ind[i]-1) * hat_mu_list[i] + rw_list[i]/action[i]) /
                       counter_ind[i] if action[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]
        counter_ind = [counter_ind[i] + 1 if action[i] >
                       0 else counter_ind[i] for i in range(MAB.K)]
        t = t + 1
        if t > T:
            return rw_T_list

    if s_remind != 0:
        action = np.zeros(MAB.K)
        for i in range(s_remind):
            action[-i-1] = 1

        add_action, _ = oracle(MAB.K, MAB.N - s_remind,
                               hat_mu_list, MAB.m_list - action)

        action = action + add_action

        rw_list = MAB.combinatorial_arm_pull(action)
        rw_T_list[t-1] = sum(rw_list)

        hat_mu_list = [((counter_ind[i]-1) * hat_mu_list[i] + rw_list[i]/action[i]) /
                       counter_ind[i] if action[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]
        counter_ind = [counter_ind[i] + 1 if action[i] >
                       0 else counter_ind[i] for i in range(MAB.K)]

        t = t + 1
        if t > T:
            return rw_T_list

    while t <= T:

        # UCB Index:
        if type == 'ucb':
            ucb_mu_list = hat_mu_list + gamma * \
                np.sqrt(1.5 * np.log(t) / counter_ind)
        elif type == 'klucb':
            ucb_mu_list = kl_ucb(counter_ind, hat_mu_list, t)

        # decision
        action, _ = oracle(MAB.K, MAB.N, hat_mu_list, MAB.m_list)

        if isawsexplore or np.random.choice([0, 1]) == 0:

            min_action = np.argmin(
                [hat_mu_list[i] if action[i] > 0 else np.inf for i in range(MAB.K)])

            action[min_action] -= 1

            explore_play, _ = oracle(
                MAB.K, 1, ucb_mu_list, MAB.m_list - action)

            action += explore_play

        # play
        rw_list = MAB.combinatorial_arm_pull(action)
        rw_T_list[t-1] = np.sum(rw_list)

        # record and update
        hat_mu_list = [((counter_ind[i]-1) * hat_mu_list[i] + rw_list[i]/min(MAB.m_list[i], action[i])
                        )/counter_ind[i] if action[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]
        counter_ind = [counter_ind[i] + 1 if action[i] >
                       0 else counter_ind[i] for i in range(MAB.K)]

        t = t + 1

    return rw_T_list


def OrchExplore(MAB, T=5000):
    ################################################################
    # Input:
    #       T: the time horizon
    # Output:
    #       a 1 x T reward list;
    ################################################################

    # Initialization
    rw_T_list = np.zeros(T)
    hat_mu_list = np.zeros(MAB.K)
    hat_nu_list = np.zeros(MAB.K)
    klucb_list = np.zeros(MAB.K)
    m_lower_list = np.ones(MAB.K, dtype=int)
    m_upper_list = np.ones(MAB.K, dtype=int) * MAB.N

    counter_ie = np.ones(MAB.K, dtype=int)
    counter_ue = np.ones(MAB.K, dtype=int)

    t = 1

    M = 10
    # initial round

    # candidate arm set
    S = np.arange(MAB.K, dtype=int)
    Y = []

    while t <= T:

        if t % 2 == 1 or Y == []:
            # PIE
            a, _ = oracle(MAB.K, MAB.N, hat_mu_list, m_lower_list)
            S = [k for k in range(MAB.K) if a[k] > 0]
            L = np.argmin(
                [hat_mu_list[k] if k in S else hat_mu_list[k]+M for k in range(MAB.K)])
            E = [k for k in range(
                MAB.K) if k not in S and klucb_list[k] > hat_mu_list[L]]
            if E != [] and np.random.uniform(0, 1) < 0.5:
                l = np.random.choice(E)
                a[L] = a[L] - 1
                a[l] = 1

            reward_list = MAB.combinatorial_arm_pull(a)
            rw_T_list[t-1] = sum(reward_list)

            hat_mu_list = [((counter_ie[i]-1) * hat_mu_list[i] + reward_list[i]/a[i]) /
                           counter_ie[i] if a[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]
            counter_ie = [counter_ie[i] + 1 if a[i] >
                          0 else counter_ie[i] for i in range(MAB.K)]

            klucb_list = kl_ucb(counter_ie, hat_mu_list, t)

        else:
            # PUE
            hat_mu_list_prime = [
                hat_mu_list[k] if k not in Y else hat_mu_list[k]+M for k in range(MAB.K)]

            a, _ = oracle(MAB.K, MAB.N, hat_mu_list_prime, m_upper_list)
            reward_list = MAB.combinatorial_arm_pull(a)
            rw_T_list[t-1] = sum(reward_list)

            hat_nu_list = [((counter_ue[i]-1) * hat_nu_list[i] + reward_list[i])/counter_ue[i]
                           if a[i] >= m_upper_list[i] else hat_nu_list[i] for i in range(MAB.K)]
            counter_ue = [counter_ue[i] + 1 if a[i] >= m_upper_list[i]
                          else counter_ue[i] for i in range(MAB.K)]

        for i in range(MAB.K):
            m_lower_list[i], m_upper_list[i] = uci(counter_ie[i], counter_ue[i], 2/T,
                                                   hat_mu_list[i], hat_nu_list[i], m_lower_list[i], m_upper_list[i])

        Y = [k for k in range(
            MAB.K) if k != L and k in S and m_lower_list[k] != m_upper_list[k]]

        t = t + 1

    return rw_T_list


def OrchExplore_with_para(MAB, xi=2, T=5000, ci_type='uci'):
    ################################################################
    # Input:
    #       T: the time horizon
    # Output:
    #       a 1 x T reward list;
    ################################################################

    # Initialization
    rw_T_list = np.zeros(T)
    hat_mu_list = np.zeros(MAB.K)
    hat_nu_list = np.zeros(MAB.K)
    klucb_list = np.zeros(MAB.K)
    m_lower_list = np.ones(MAB.K, dtype=int)
    m_upper_list = np.ones(MAB.K, dtype=int) * MAB.N

    counter_ie = np.ones(MAB.K, dtype=int)
    counter_ue = np.ones(MAB.K, dtype=int)

    t = 1

    M = 10
    # initial round

    # candidate arm set
    S = np.arange(MAB.K, dtype=int)
    Y = []

    while t <= T:

        if t % 2 == 1 or Y == []:
            # PIE
            a, _ = oracle(MAB.K, MAB.N, hat_mu_list, m_lower_list)
            S = [k for k in range(MAB.K) if a[k] > 0]
            L = np.argmin(
                [hat_mu_list[k] if k in S else hat_mu_list[k]+M for k in range(MAB.K)])
            E = [k for k in range(
                MAB.K) if k not in S and klucb_list[k] > hat_mu_list[L]]
            if E != [] and np.random.uniform(0, 1) < 0.5:
                l = np.random.choice(E)
                a[L] = a[L] - 1
                a[l] = 1

            reward_list = MAB.combinatorial_arm_pull(a)
            rw_T_list[t-1] = sum(reward_list)

            hat_mu_list = [((counter_ie[i]-1) * hat_mu_list[i] + reward_list[i]/a[i]) /
                           counter_ie[i] if a[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]
            counter_ie = [counter_ie[i] + 1 if a[i] >
                          0 else counter_ie[i] for i in range(MAB.K)]

            klucb_list = kl_ucb(counter_ie, hat_mu_list, t)

        else:
            # PUE
            hat_mu_list_prime = [
                hat_mu_list[k] if k not in Y else hat_mu_list[k]+M for k in range(MAB.K)]

            a, _ = oracle(MAB.K, MAB.N, hat_mu_list_prime, m_upper_list)
            reward_list = MAB.combinatorial_arm_pull(a)
            rw_T_list[t-1] = sum(reward_list)

            hat_nu_list = [((counter_ue[i]-1) * hat_nu_list[i] + reward_list[i])/counter_ue[i]
                           if a[i] >= m_upper_list[i] else hat_nu_list[i] for i in range(MAB.K)]
            counter_ue = [counter_ue[i] + 1 if a[i] >= m_upper_list[i]
                          else counter_ue[i] for i in range(MAB.K)]

        for i in range(MAB.K):

            if ci_type == 'uci':
                m_lower_list[i], m_upper_list[i] = uci(counter_ie[i], counter_ue[i], xi/T,
                                                       hat_mu_list[i], hat_nu_list[i], m_lower_list[i], m_upper_list[i])
            elif ci_type == 'hfd':
                m_lower_list[i], m_upper_list[i] = hfd_uci(counter_ie[i], counter_ue[i], xi/(T**2),
                                                           hat_mu_list[i], hat_nu_list[i], m_lower_list[i], m_upper_list[i])

        Y = [k for k in range(
            MAB.K) if k != L and k in S and m_lower_list[k] != m_upper_list[k]]

        t = t + 1

    return rw_T_list


def OrchExplore_easy(MAB, T=5000):
    ################################################################
    # Input:
    #       T: the time horizon
    # Output:
    #       a 1 x T reward list;
    ################################################################

    # Initialization
    rw_T_list = np.zeros(T)
    hat_mu_list = np.zeros(MAB.K)
    hat_nu_list = np.zeros(MAB.K)
    klucb_list = np.zeros(MAB.K)
    m_lower_list = np.ones(MAB.K, dtype=int)
    m_upper_list = np.ones(MAB.K, dtype=int) * MAB.N

    counter_ie = np.ones(MAB.K, dtype=int)
    counter_ue = np.ones(MAB.K, dtype=int)

    t = 1

    M = 10
    # initial round

    # candidate arm set
    S = np.arange(MAB.K, dtype=int)
    Y = []

    while t <= T:

        if t % 2 == 1 or Y == []:
            # PIE
            a, _ = oracle(MAB.K, MAB.N, hat_mu_list, m_lower_list)
            S = [k for k in range(MAB.K) if a[k] > 0]
            L = np.argmin(
                [hat_mu_list[k] if k in S else hat_mu_list[k]+M for k in range(MAB.K)])
            E = [k for k in range(
                MAB.K) if k not in S and klucb_list[k] > hat_mu_list[L]]
            if E != [] and np.random.uniform(0, 1) < 0.5:
                l = np.random.choice(E)
                a[L] = a[L] - 1
                a[l] = 1

            reward_list = MAB.combinatorial_arm_pull(a)
            rw_T_list[t-1] = sum(reward_list)

            hat_mu_list = [((counter_ie[i]-1) * hat_mu_list[i] + reward_list[i]/a[i]) /
                           counter_ie[i] if a[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]
            counter_ie = [counter_ie[i] + 1 if a[i] >
                          0 else counter_ie[i] for i in range(MAB.K)]

            klucb_list = kl_ucb(counter_ie, hat_mu_list, t)

        else:
            # PUE
            hat_mu_list_prime = [
                hat_mu_list[k] if k not in Y else hat_mu_list[k]+M for k in range(MAB.K)]

            a, _ = oracle(MAB.K, MAB.N, hat_mu_list_prime, m_upper_list)
            reward_list = MAB.combinatorial_arm_pull(a)
            rw_T_list[t-1] = sum(reward_list)

            hat_nu_list = [((counter_ue[i]-1) * hat_nu_list[i] + reward_list[i])/counter_ue[i]
                           if a[i] >= m_upper_list[i] else hat_nu_list[i] for i in range(MAB.K)]
            counter_ue = [counter_ue[i] + 1 if a[i] >= m_upper_list[i]
                          else counter_ue[i] for i in range(MAB.K)]

        for i in range(MAB.K):
            m_lower_list[i], m_upper_list[i] = uci_easy(counter_ie[i], counter_ue[i], 2/T,
                                                        hat_mu_list[i], hat_nu_list[i], MAB.N)

        Y = [k for k in range(
            MAB.K) if k != L and k in S and m_lower_list[k] != m_upper_list[k]]

        t = t + 1

    return rw_T_list


def MPTS(MAB, T=5000):
    ################################################################
    # Input:
    #       t0: 1-1 player - arm exploration
    #       t1: N-1 player - arm exploitation
    #       T: the time horizon
    # Output:
    #       a 1 x T reward list;
    ################################################################

    # Initialization
    rw_T_list = np.zeros(T)

    t = 1

    positive_ind = np.ones(MAB.K, dtype=int)
    negative_ind = np.ones(MAB.K, dtype=int)

    while t <= T:

        # UCB Index:
        ts_index = [np.random.beta(positive_ind[i], negative_ind[i])
                    for i in range(MAB.K)]

        # decision
        action, _ = oracle(MAB.K, MAB.N, ts_index, MAB.m_list)

        # play
        rw_list = MAB.combinatorial_arm_pull(action)
        rw_T_list[t-1] = np.sum(rw_list)

        positive_ind = [positive_ind[i] + 1 if rw_list[i]
                        > 0 else positive_ind[i] for i in range(MAB.K)]

        negative_ind = [negative_ind[i] + 1 if action[i] >
                        0 and rw_list[i] == 0 else negative_ind[i] for i in range(MAB.K)]

        t = t + 1

    return rw_T_list


def MPTS_Gaussian(MAB, T=5000):
    ################################################################
    # Input:
    #       t0: 1-1 player - arm exploration
    #       t1: N-1 player - arm exploitation
    #       T: the time horizon
    # Output:
    #       a 1 x T reward list;
    ################################################################

    # Initialization
    rw_T_list = np.zeros(T)

    t = 1

    hat_mu_list = np.zeros(MAB.K)
    counter_list = np.ones(MAB.K, dtype=int)

    while t <= T:

        # UCB Index:
        ts_index = [np.random.normal(loc=hat_mu_list[i], scale=np.sqrt(0.5/(counter_list[i])))
                    for i in range(MAB.K)]

        # decision
        action, _ = oracle(MAB.K, MAB.N, ts_index, MAB.m_list)

        # play
        rw_list = MAB.combinatorial_arm_pull(action)
        rw_T_list[t-1] = np.sum(rw_list)

        hat_mu_list = [((counter_list[i]-1) * hat_mu_list[i] + rw_list[i]/action[i]) /
                       counter_list[i] if action[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]

        counter_list = [counter_list[i] + 1 if action[i] >
                        0 else counter_list[i] for i in range(MAB.K)]

        t = t + 1

    return rw_T_list


# create algorithms for Combinatorial Action Space, size combinatorial(N+K-1, K-1) \sim O(K^N) for fixed N and K \to \infinity
def SE_com(MAB, gamma=1, T=5000):
    ################################################################
    # Input: Assume Known m
    #       T: the time horizon
    # Output:
    #       a 1 x T reward list;
    ################################################################

    action_space = list(partition(MAB.K, MAB.N))
    action_size = len(action_space)

    # Initialization
    rw_T_list = np.zeros(T)
    hat_mu_list = np.zeros(action_size)

    tau = 1
    S = np.arange(action_size, dtype=int)
    t = 1
    gma = gamma

    while t <= T:
        order = np.argsort([hat_mu_list[i] for i in S])
        s = len(S)
        S_prime = S

        if s == 1:

            action_index = S[0]

            rw = sum(MAB.combinatorial_arm_pull(action_space[action_index]))
            rw_T_list[t-1] = rw

            t = t + 1
            if t > T:
                return rw_T_list

            hat_mu_list[action_index] = (
                (tau-1) * hat_mu_list[action_index] + rw)/tau

        else:
            gap = gma * np.sqrt(MAB.K) * U(tau, T)

            for i in range(s-1, 0, -1):
                if hat_mu_list[S_prime[order[-1]]] - hat_mu_list[S_prime[order[-i-1]]] > gap:
                    S = masked_del(S, S_prime[order[-i-1]])
                else:
                    break

            # renew s after elimination
            s = len(S)

            for action_index in S:

                rw = sum(MAB.combinatorial_arm_pull(
                    action_space[action_index]))
                rw_T_list[t-1] = rw
                t = t + 1
                if t > T:
                    return rw_T_list

                hat_mu_list[action_index] = (
                    (tau-1) * hat_mu_list[action_index] + rw)/tau

        tau = tau + 1

    return rw_T_list


def UCB_com(MAB, gamma=1, T=5000):
    ################################################################
    # Input:
    #       t0: 1-1 player - arm exploration
    #       t1: N-1 player - arm exploitation
    #       T: the time horizon
    # Output:
    #       a 1 x T reward list;
    ################################################################

    action_space = list(partition(MAB.K, MAB.N))
    action_size = len(action_space)

    # Initialization
    rw_T_list = np.zeros(T)
    hat_mu_list = np.zeros(action_size)

    t = 1

    counter_ind = np.ones(action_size, dtype=int)

    for action_index in range(action_size):

        rw = sum(MAB.combinatorial_arm_pull(action_space[action_index]))
        rw_T_list[t-1] = rw

        hat_mu_list[action_index] = (
            (counter_ind[action_index]-1) * hat_mu_list[action_index] + rw)/counter_ind[action_index]

        counter_ind[action_index] += 1

        t = t + 1
        if t > T:
            return rw_T_list

    while t <= T:

        # UCB Index:
        ucb_mu_list = hat_mu_list + gamma * \
            np.sqrt(MAB.K * 1.5 * np.log(t) / counter_ind)

        # decision
        action_index = np.argmax(ucb_mu_list)

        # play
        rw = sum(MAB.combinatorial_arm_pull(action_space[action_index]))
        rw_T_list[t-1] = rw

        # record and update
        hat_mu_list[action_index] = (
            (counter_ind[action_index]-1) * hat_mu_list[action_index] + rw)/counter_ind[action_index]

        counter_ind[action_index] += 1

        t = t + 1

    return rw_T_list


def TS_com(MAB, T=5000):
    ################################################################
    # Input:
    #       t0: 1-1 player - arm exploration
    #       t1: N-1 player - arm exploitation
    #       T: the time horizon
    # Output:
    #       a 1 x T reward list;
    ################################################################

    action_space = list(partition(MAB.K, MAB.N))
    action_size = len(action_space)

    # Initialization
    rw_T_list = np.zeros(T)

    t = 1

    ts_stat = np.ones([action_size, MAB.N+1], dtype=int)

    weight = np.arange(MAB.N+1)

    while t <= T:

        # Sampling
        ts_mu_list = [np.dot(weight, np.random.dirichlet(
            ts_stat[i, :])) for i in range(action_size)]

        # decision
        action_index = np.argmax(ts_mu_list)

        # print(t, action_space[action_index])

        # play
        rw = sum(MAB.combinatorial_arm_pull(action_space[action_index]))
        rw_T_list[t-1] = rw

        ts_stat[action_index, int(rw)] += 1

        t = t + 1

    return rw_T_list


def TS_com_Gaussian(MAB, T=5000):
    ################################################################
    # Input:
    #       t0: 1-1 player - arm exploration
    #       t1: N-1 player - arm exploitation
    #       T: the time horizon
    # Output:
    #       a 1 x T reward list;
    ################################################################

    action_space = list(partition(MAB.K, MAB.N))
    action_size = len(action_space)

    # Initialization
    rw_T_list = np.zeros(T)

    t = 1

    hat_mu_list = np.zeros(action_size)
    counter_list = np.ones(action_size, dtype=int)

    while t <= T:

        # Sampling
        ts_mu_list = [np.random.normal(loc=hat_mu_list[i], scale=np.sqrt(
            MAB.N/2 / (counter_list[i]))) for i in range(action_size)]

        # decision
        action_index = np.argmax(ts_mu_list)

        # print(t, action_space[action_index])

        # play
        rw = sum(MAB.combinatorial_arm_pull(action_space[action_index]))
        rw_T_list[t-1] = rw

        hat_mu_list[action_index] = (
            (counter_list[action_index]-1) * hat_mu_list[action_index] + rw) / counter_list[action_index]
        counter_list[action_index] += 1

        t = t + 1

    return rw_T_list


def ETC_CUCB(MAB, gamma=1, xi=0.01, T=5000, type='uci'):
    ################################################################
    # Input:
    #       t0: 1-1 player - arm exploration
    #       t1: N-1 player - arm exploitation
    #       T: the time horizon
    # Output:
    #       a 1 x T reward list;
    ################################################################

    # Initialization
    rw_T_list = np.zeros(T)
    hat_mu_list = np.zeros(MAB.K)
    hat_nu_list = np.zeros(MAB.K)
    # m_flag = np.zeros(MAB.K)

    hat_m_list = np.ones(MAB.K, dtype=int)
    hat_m_upper = np.ones(MAB.K, dtype=int) * MAB.N

    # time index and sample size index
    tau_list = np.ones(MAB.K, dtype=int)
    iota_list = np.ones(MAB.K, dtype=int)
    t = 1

    # auxillary for ETC
    S = np.arange(MAB.K, dtype=int)

    # 1st phase: ETC exploration
    while len(S) > 0:

        # one round of IE
       # individually explore S and update hat_mu_list
        s = len(S)
        s_round = s // MAB.N
        s_remind = s % MAB.N

        for i in range(s_round):
            action = np.zeros(MAB.K)
            for j in range(MAB.N):
                action[S[j + i*MAB.N]] = 1

            rw_list = MAB.combinatorial_arm_pull(action)
            rw_T_list[t-1] = sum(rw_list)

            hat_mu_list = [((tau_list[i]-1) * hat_mu_list[i] + (rw_list[i]/action[i])) /
                           tau_list[i] if action[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]
            t = t + 1
            if t > T:
                return rw_T_list

        if s_remind != 0:
            action = np.zeros(MAB.K)
            for i in range(s_remind):
                action[S[-i-1]] = 1

            add_action, _ = oracle(
                MAB.K, MAB.N - s_remind, hat_mu_list, hat_m_list - action)

            action = action + add_action

            rw_list = MAB.combinatorial_arm_pull(action)
            rw_T_list[t-1] = sum(rw_list)

            hat_mu_list = [((tau_list[i]-1) * hat_mu_list[i] + (rw_list[i]/action[i])) /
                           tau_list[i] if action[i] > 0 else hat_mu_list[i] for i in range(MAB.K)]
            t = t + 1
            if t > T:
                return rw_T_list

        tau_list = [tau_list[i]+1 if i in S else tau_list[i]
                    for i in range(MAB.K)]

        # one round of UE
        S_prime = S
        for arm in S_prime:
            action = np.zeros(MAB.K)
            action[arm] = hat_m_upper[arm]

            add_action, _ = oracle(MAB.K, MAB.N - action[arm], hat_mu_list, [
                0 if i == arm else hat_m_list[i] for i in range(MAB.K)])
            action = action + add_action

            rw_list = MAB.combinatorial_arm_pull(action)
            rw_T_list[t-1] = sum(rw_list)

            hat_nu_list[arm] = (
                (iota_list[arm]-1) * hat_nu_list[arm] + rw_list[arm])/iota_list[arm]

            t = t+1
            if t > T:
                return rw_T_list

            iota_list[arm] = iota_list[arm] + 1

            if type == 'uci':
                hat_m_list[arm], hat_m_upper[arm] = uci(
                    tau_list[arm], iota_list[arm], xi/T, hat_mu_list[arm], hat_nu_list[arm], hat_m_list[arm], hat_m_upper[arm])
            elif type == 'hfd':
                hat_m_list[arm], hat_m_upper[arm] = hfd_uci(
                    tau_list[arm], iota_list[arm], xi/(T**2), hat_mu_list[arm], hat_nu_list[arm], hat_m_list[arm], hat_m_upper[arm])

            if hat_m_list[arm] == hat_m_upper[arm]:
                S = masked_del(S, arm)

        # print(
        #     f"candidate arm set ={S}, time={t, tau_list, iota_list}, hat_mu_list={hat_mu_list} hat_m_list={hat_m_list}, hat_m_upper={hat_m_upper}")

    # 2nd phase: UCB
    while t <= T:

        # UCB Index:
        ucb_mu_list = hat_mu_list + gamma * \
            np.sqrt(1.5 * np.log(t) / tau_list)

        # decision
        action, _ = oracle(MAB.K, MAB.N, ucb_mu_list, hat_m_list)

        # play
        rw_list = MAB.combinatorial_arm_pull(action)
        rw_T_list[t-1] = np.sum(rw_list)

        # record and update
        for i in range(MAB.K):
            if action[i] > 0:
                hat_mu_list[i] = (
                    (tau_list[i]-1) * hat_mu_list[i] + rw_list[i]/action[i])/tau_list[i]
                tau_list[i] += 1

        t = t + 1

    return rw_T_list

